using System;
using System.Threading;
using System.Reflection;
using System.Reflection.Emit;

public class Counter
{
    public int CountLoop(int totalCount)
    {
        int total = 0;

        for(int loopIndex = 1; loopIndex <= totalCount; loopIndex++)
        {
            total++;
        }
        return total;
    }
}

public class Reflector
{
    ICounter counter = null;

    public int CountSimple(int totalCount)
    {
        if (counter == null)
        {
           AssemblyBuilder assemblyBuilder = Emitter(totalCount);
           counter = (ICounter) assemblyBuilder.CreateInstance("Counter");
        }

        return (counter.GeneratedCountSimple());
    }

    AssemblyBuilder Emitter(int totalCount)
    {
        AssemblyName assemblyName = new AssemblyName();
        assemblyName.Name = "CounterAssembly";

        AssemblyBuilder assemblyBuilder = Thread.GetDomain().DefineDynamicAssembly(
            assemblyName, AssemblyBuilderAccess.Run);

        ModuleBuilder moduleBuilder = assemblyBuilder.DefineDynamicModule("Module1");

        TypeBuilder typeBuilder = moduleBuilder.DefineType("Counter", TypeAttributes.Public);

        typeBuilder.AddInterfaceImplementation(typeof(ICounter));

        Type[] paramTypes = new Type[0];
        Type returnType = typeof(System.Int32);

        MethodBuilder generatedMethod =
            typeBuilder.DefineMethod("GeneratedCountSimple", MethodAttributes.Virtual | 
            MethodAttributes.Public, returnType, paramTypes);

        ILGenerator ilGenerator = generatedMethod.GetILGenerator();
        ilGenerator.Emit(OpCodes.Ldc_I4, 0);

        for (int loopIndex = 1; loopIndex <= totalCount; loopIndex++)
        {
            ilGenerator.Emit(OpCodes.Ldc_I4, 1);
            ilGenerator.Emit(OpCodes.Add);
        }

        ilGenerator.Emit(OpCodes.Ret);

        MethodInfo methodInfo = typeof(ICounter).GetMethod("GeneratedCountSimple");

        typeBuilder.DefineMethodOverride(generatedMethod, methodInfo);
        typeBuilder.CreateType();

        return assemblyBuilder;
    }
}

public interface ICounter
{
    int GeneratedCountSimple();
}

public class ch14_09
{
    public static void Main()
    {
        int total = 0;

        Counter counter = new Counter(); 

        DateTime begin = DateTime.Now;            
        for (int loopIndex = 0; loopIndex < 10000000; loopIndex++)
        {
            total = counter.CountLoop(30);
        }
        TimeSpan duration = DateTime.Now - begin;

        System.Console.WriteLine("Looped to {0} 10,000,000 times in {1} milliseconds.", 
            total, duration.TotalMilliseconds);

        Reflector reflector = new Reflector();

        begin = DateTime.Now; 
        for (int loopIndex = 0; loopIndex < 10000000; loopIndex++)
        {
            total = reflector.CountSimple(30);
        }
        duration = DateTime.Now - begin;

        System.Console.WriteLine("Counted to {0} 10,000,000 times in {1} milliseconds.", 
            total, duration.TotalMilliseconds);
    }
}
